using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;

using Emgu.CV;
using Emgu.CV.Structure;
using Emgu.CV.UI;

namespace TestProject
{
    /// <summary>
    ///  
    /// </summary>
    public partial class MainForm : Form
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 
        /// </summary>
        private StopSignDetector detector;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region  - MainForm()

        /// <summary>
        /// 
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            using(Image<Bgr, Byte> modelImage = new Image<Bgr, Byte>("stop-sign-model.png"))
            {
                using (Image<Bgr, Byte> image = new Image<Bgr, byte>("stop-sign.jpg"))
                {
                    this.detector = new StopSignDetector(modelImage);

                    ProcessImage(image);
                }
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region ̹ ε ư Ŭ óϱ - loadImageButton_Click(sender, e)

        /// <summary>
        /// ̹ ε ư Ŭ óϱ
        /// </summary>
        /// <param name="sender">̺Ʈ ߻</param>
        /// <param name="e">̺Ʈ </param>
        private void loadImageButton_Click(object sender, EventArgs e)
        {
            DialogResult result = openFileDialog.ShowDialog();

            if(result == DialogResult.OK)
            {
                Image<Bgr, Byte> image;

                try
                {
                    image = new Image<Bgr, byte>(openFileDialog.FileName);
                }
                catch
                {
                    MessageBox.Show("ϴ   ƴմϴ.");

                    return;
                }

                ProcessImage(image);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region ̺  ̹ ߰ϱ - AddLabelAndImage(startPoint, text, image)

        /// <summary>
        /// ̺  ̹ ߰ϱ
        /// </summary>
        /// <param name="startPoint"> ġ</param>
        /// <param name="text">ؽƮ</param>
        /// <param name="image">̹</param>
        private void AddLabelAndImage(ref Point startPoint, string text, IImage image)
        {
            Label label = new Label();

            this.resultPanel.Controls.Add(label);

            label.Width    = 100;
            label.Height   = 30;
            label.Location = startPoint;
            label.Text     = text;

            startPoint.Y += label.Height;

            ImageBox imageBox = new ImageBox();

            this.resultPanel.Controls.Add(imageBox);

            imageBox.ClientSize = image.Size;
            imageBox.Location   = startPoint;
            imageBox.Image      = image;

            startPoint.Y += imageBox.Height + 10;
        }

        #endregion
        #region ̹ óϱ - ProcessImage(image)

        /// <summary>
        /// ̹ óϱ
        /// </summary>
        /// <param name="image">̹</param>
        private void ProcessImage(Image<Bgr, byte> image)
        {
            Stopwatch stopwatch = Stopwatch.StartNew();

            List<Image<Gray, Byte>> imageList     = new List<Image<Gray, byte>>();
            List<Rectangle>         rectangleList = new List<Rectangle>();

            this.detector.DetectStopSign(image, imageList, rectangleList);

            stopwatch.Stop();

            this.messageLabel.Text = String.Format("STOP ǥ  ð : {0} и", stopwatch.Elapsed.TotalMilliseconds);

            this.resultPanel.Controls.Clear();

            Point startPoint = new Point(10, 10);

            for(int i = 0; i < imageList.Count; i++)
            {
                Rectangle rectangle = rectangleList[i];

                AddLabelAndImage
                (
                    ref startPoint,
                    string.Format
                    (
                        "STOP ǥ [{0}, {1}]",
                        rectangle.Location.Y + rectangle.Width  / 2,
                        rectangle.Location.Y + rectangle.Height / 2
                    ),
                    imageList[i]
                );

                image.Draw(rectangle, new Bgr(Color.Aquamarine), 2);
            }

            this.imageBox.Image = image;
        }

        #endregion
    }
}