using System;
using System.Collections.Generic;
using System.Drawing;

using Emgu.CV;
using Emgu.CV.CvEnum;
using Emgu.CV.Features2D;
using Emgu.CV.Structure;
using Emgu.Util;

namespace TestProject
{
    /// <summary>
    /// STOP ǥ 
    /// </summary>
    public class StopSignDetector : DisposableObject
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 
        /// </summary>
        private Features2DTracker<float> tracker;

        /// <summary>
        /// 
        /// </summary>
        private SURFDetector detector;

        /// <summary>
        /// Ȱ 
        /// </summary>
        private MemStorage octagonStorage;

        /// <summary>
        /// Ȱ 
        /// </summary>
        private Contour<Point> octagonContour;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region  - StopSignDetector(modelImage)

        /// <summary>
        /// 
        /// </summary>
        /// <param name="modelImage"> ̹</param>
        public StopSignDetector(Image<Bgr, Byte> modelImage)
        {
            this.detector = new SURFDetector(500, false);

            using(Image<Gray, Byte> maskImage = GetRedPixelMaskImage(modelImage))
            {
                ImageFeature<float>[] imageFeatureArray = this.detector.DetectFeatures(maskImage, null);

                if(imageFeatureArray.Length == 0)
                {
                    throw new Exception("No image feature has been found in the stop sign model");
                }

                this.tracker = new Features2DTracker<float>(imageFeatureArray);  
            }

            this.octagonStorage = new MemStorage();

            this.octagonContour = new Contour<Point>(this.octagonStorage);

            this.octagonContour.PushMulti
            (
                new Point[]
                { 
                    new Point(1, 0),
                    new Point(2, 0),
                    new Point(3, 1),
                    new Point(3, 2),
                    new Point(2, 3),
                    new Point(1, 3),
                    new Point(0, 2),
                    new Point(0, 1)
                },
                BACK_OR_FRONT.FRONT
            );
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region STOP ǥ ϱ - DetectStopSign(sourceImage, imageList, rectangleList)

        /// <summary>
        /// STOP ǥ ϱ
        /// </summary>
        /// <param name="sourceImage">ҽ ̹</param>
        /// <param name="imageList">̹ Ʈ</param>
        /// <param name="rectangleList">簢 Ʈ</param>
        public void DetectStopSign(Image<Bgr, byte> sourceImage, List<Image<Gray, Byte>> imageList, List<Rectangle> rectangleList)
        {
            Image<Bgr, Byte> smoothImage = sourceImage.SmoothGaussian(5, 5, 1.5, 1.5);

            Image<Gray, Byte> maskImage = GetRedPixelMaskImage(smoothImage);

            maskImage._Dilate(1);

            maskImage._Erode(1);

            using(Image<Gray, Byte> cannyImage = maskImage.Canny(100, 50))
            {
                using(MemStorage storage = new MemStorage())
                {
                    Contour<Point> pointContour = cannyImage.FindContours
                    (
                        CHAIN_APPROX_METHOD.CV_CHAIN_APPROX_SIMPLE,
                        RETR_TYPE.CV_RETR_TREE,
                        storage
                    );

                    FindStopSign(sourceImage, imageList, rectangleList, pointContour);
                }
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region ü ҽ ϱ - DisposeObject()

        /// <summary>
        /// ü ҽ ϱ
        /// </summary>
        protected override void DisposeObject()
        {
            this.tracker.Dispose();

            this.octagonStorage.Dispose();
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region  ȼ ũ ̹ ϱ - GetRedPixelMaskImage(sourceImage)

        /// <summary>
        ///  ȼ ũ ̹ ϱ
        /// </summary>
        /// <param name="sourceImage">ҽ ̹</param>
        /// <returns> ȼ ũ ̹</returns>
        private Image<Gray, Byte> GetRedPixelMaskImage(Image<Bgr, byte> sourceImage)
        {
            using(Image<Hsv, Byte> hsvImage = sourceImage.Convert<Hsv, Byte>())
            {
                Image<Gray, Byte>[] channelImageArray = hsvImage.Split();

                try
                {
                    CvInvoke.cvInRangeS
                    (
                        channelImageArray[0],
                        new MCvScalar(20),
                        new MCvScalar(160),
                        channelImageArray[0]
                    );

                    channelImageArray[0]._Not();

                    channelImageArray[1]._ThresholdBinary(new Gray(10), new Gray(255.0));

                    CvInvoke.cvAnd(channelImageArray[0], channelImageArray[1], channelImageArray[0], IntPtr.Zero);
                }
                finally
                {
                    channelImageArray[1].Dispose();
                    channelImageArray[2].Dispose();
                }
    
                return channelImageArray[0];
            }
        }

        #endregion
        #region STOP ǥ ã - FindStopSign(sourceImage, imageList, rectangleList, pointContour)

        /// <summary>
        /// STOP ǥ ã
        /// </summary>
        /// <param name="sourceImage">ҽ ̹</param>
        /// <param name="imageList">̹ Ʈ</param>
        /// <param name="rectangleList">簢 Ʈ</param>
        /// <param name="pointContour">Ʈ </param>
        private void FindStopSign(Image<Bgr, byte> sourceImage, List<Image<Gray, Byte>> imageList, List<Rectangle> rectangleList, Contour<Point> pointContour)
        {
            for(; pointContour != null; pointContour = pointContour.HNext)
            {
                pointContour.ApproxPoly(pointContour.Perimeter * 0.02, 0, pointContour.Storage);

                if(pointContour.Area > 200)
                {
                    double ratio = CvInvoke.cvMatchShapes
                    (
                        this.octagonContour,
                        pointContour,
                        CONTOURS_MATCH_TYPE.CV_CONTOURS_MATCH_I3,
                        0
                    );

                    if(ratio > 0.1)
                    {
                        Contour<Point> childPointContour = pointContour.VNext;

                        if(childPointContour != null)
                        {
                            FindStopSign(sourceImage, imageList, rectangleList, childPointContour);
                        }

                        continue;
                    }

                    Rectangle rectangle = pointContour.BoundingRectangle;

                    Image<Gray, Byte> candidateImage;

                    using(Image<Bgr, Byte> temporaryImage = sourceImage.Copy(rectangle))
                    {
                        candidateImage = temporaryImage.Convert<Gray, byte>();
                    }

                    using(Image<Gray, Byte> maskImage = new Image<Gray, byte>(rectangle.Size))
                    {
                        maskImage.Draw
                        (
                            pointContour,
                            new Gray(255),
                            new Gray(255),
                            0,
                            -1,
                            new Point(-rectangle.X, -rectangle.Y)
                        );

                        double mean = CvInvoke.cvAvg(candidateImage, maskImage).v0;

                        candidateImage._ThresholdBinary(new Gray(mean), new Gray(255.0));

                        candidateImage._Not();

                        maskImage._Not();

                        candidateImage.SetValue(0, maskImage);
                    }

                    ImageFeature<float>[] imageFeatureArray = this.detector.DetectFeatures(candidateImage, null);

                    int minimumMatchCount = 10;

                    if(imageFeatureArray != null && imageFeatureArray.Length >= minimumMatchCount)
                    {
                        Features2DTracker<float>.MatchedImageFeature[] matchedImageFeatureArray = this.tracker.MatchFeature(imageFeatureArray, 2);

                        int goodMatchCount = 0;

                        foreach(Features2DTracker<float>.MatchedImageFeature matchedImageFeature in matchedImageFeatureArray)
                        {
                            if(matchedImageFeature.SimilarFeatures[0].Distance < 0.5)
                            {
                                goodMatchCount++;
                            }
                        }

                        if(goodMatchCount >= minimumMatchCount)
                        {
                            rectangleList.Add(rectangle);

                            imageList.Add(candidateImage);
                        }
                    }
                }
            }
        }

        #endregion
    }
}